#!/usr/bin/env python
"""
patch.py - Minecraft Log4j patching script.
By ~keith. This script was written in a couple hours, so it's janky as
fuck. I promise I don't normally write code this bad.
"""
import os
import shutil
import platform
import zipfile
import json
import hashlib
import time
import datetime
import secrets

BAK_EXT = ".{}.bak".format(int(time.time()))

def bold(text):
	return "\x1b[1m" + text + "\x1b[0m"

def prompt_yn(msg):
	while True:
		text = input("{0} (y or n): ".format(msg)).strip().lower()
		if text.startswith("y"):
			return True
		elif text.startswith("n"):
			return False
		else:
			print("Please type y or n. (Or press ^C to abort.)")

def prompt_value(msg, default):
	text = input("{0} ({1}): ".format(msg, default)).strip()
	if len(text) > 0:
		return text
	return default

MINECRAFT_ROOT = os.path.expanduser("~/.minecraft/")

ENABLE_SELECTION = True
ALWAYS_PATCH_COPY = True

NEW_LOG4J_VERSION = "2.17.1"
NEW_LOG4J_ROOT = "apache-log4j-2.17.1-bin/"
LOG4J_REPLACE = {
	"org.apache.logging.log4j:log4j-core": "log4j-core-2.17.1.jar",
	"org.apache.logging.log4j:log4j-api": "log4j-api-2.17.1.jar",
	"org.apache.logging.log4j:log4j-slf4j18-impl": "log4j-slf4j18-impl-2.17.1.jar",
}
CHANGE_VERSION = True

def get_mc_versions():
	with open(os.path.join(MINECRAFT_ROOT, "launcher_profiles.json"), 'r') as fh:
		launcher_cfg = json.load(fh)

	with open(os.path.join(MINECRAFT_ROOT, "versions/version_manifest_v2.json"), 'r') as fh:
		latests = json.load(fh)['latest']

	# Offer to disable analytics
	try:
		if launcher_cfg['settings']['enableAnalytics']:
			print(bold("Warning:"), "Launcher analytics enabled")
			print("Analytics are harmful to your privacy, security, and freedom.")
			print("Would you like to disable them? This won't affect your game.")
			print()
			if prompt_yn("Disable launcher analytics?"):
				launcher_cfg['settings']['enableAnalytics'] = False
				with open(os.path.join(MINECRAFT_ROOT, "launcher_profiles.json"), 'w') as fh:
					json.dump(launcher_cfg, fh, indent="\t")
			print()
	except:
		pass

	versions = []
	for profile in launcher_cfg['profiles'].values():
		ver = profile['lastVersionId']
		if ver == 'latest-snapshot':
			ver = latests['snapshot']
		elif ver == 'latest-release':
			ver = latests['release']

		if ver not in versions:
			versions.append(ver)
	return versions

def get_dependencies(version):
	with open(os.path.join(MINECRAFT_ROOT, "versions/{0}/{0}.json".format(version)), 'r') as fh:
		version_info = json.load(fh)
	deps = {}
	for library in version_info['libraries']:
		name, _, version = library['name'].rpartition(':')
		deps[name] = {
			'version': version,
			'path': library['downloads']['artifact']['path'],
			'sha1': library['downloads']['artifact']['sha1'],
			'size': library['downloads']['artifact']['size']
		}
	return deps


def patch_jar(dep_path, dep_name):
	jar_path = os.path.join(MINECRAFT_ROOT, "libraries", dep_path)
	assert os.path.exists(jar_path), "patch_jar({}) failed, file not found".format(dep_path)

	shutil.move(jar_path, jar_path + BAK_EXT)

	new_jar_path = os.path.join(NEW_LOG4J_ROOT, LOG4J_REPLACE[dep_name])
	
	if CHANGE_VERSION:
		version_dir = os.path.normpath(os.path.join(jar_path, "../../", NEW_LOG4J_VERSION))
		if not os.path.isdir(version_dir):
			os.mkdir(version_dir)
		jar_path = os.path.join(version_dir, LOG4J_REPLACE[dep_name])
		if not os.path.exists(jar_path):
			shutil.copy2(new_jar_path, jar_path)
	else:
		shutil.copy2(new_jar_path, jar_path)
	
	return os.path.relpath(jar_path, os.path.join(MINECRAFT_ROOT, "libraries"))

def hash_file(path):
	h = hashlib.sha1()
	with open(path, 'rb') as fh:
		while len(data := fh.read(4096)) > 0:
			h.update(data)
	return h.hexdigest()

def patch_deps(version, patched_libs, paths):
	version_path = os.path.join(MINECRAFT_ROOT, "versions/{0}/{0}.json".format(version))

	shutil.copy2(version_path, version_path + BAK_EXT)

	with open(version_path, 'r') as fh:
		version_info = json.load(fh)

	version_info['id'] = version

	for library in version_info['libraries']:
		if library['name'] in patched_libs:
			if library['name'] in paths:
				library['downloads']['artifact']['path'] = paths[library['name']]
			
			if CHANGE_VERSION:
				library['name'] = library['name'].rpartition(':')[0] + ":" + NEW_LOG4J_VERSION
			
			path = os.path.join(MINECRAFT_ROOT, "libraries", library['downloads']['artifact']['path'])
			print("updt {}".format(library['name']))
			print(" --> {}".format(path))

			library['downloads']['artifact']['url'] = ""
			library['downloads']['artifact']['sha1'] = hash_file(path)
			library['downloads']['artifact']['size'] = os.stat(path).st_size

	with open(version_path, 'w') as fh:
		json.dump(version_info, fh, indent="\t")

def copy_version(version):
	version_path = os.path.join(MINECRAFT_ROOT, "versions/{}".format(version))
	patched_path = os.path.join(MINECRAFT_ROOT, "versions/{}-patched".format(version))
	shutil.copytree(version_path, patched_path, dirs_exist_ok=True)

	shutil.move(os.path.join(patched_path, version + ".json"), os.path.join(patched_path, version + "-patched.json"))
	shutil.move(os.path.join(patched_path, version + ".jar"), os.path.join(patched_path, version + "-patched.jar"))

def check_manifest(versions):
	# Check the manifest for versions we can't modify.
	# And make -patched copies of their directories.

	with open(os.path.join(MINECRAFT_ROOT, "versions/version_manifest_v2.json"), 'r') as fh:
		manifest = json.load(fh)

	protected = []
	for version_info in manifest['versions']:
		if version_info['id'] in versions:
			version = version_info['id']
			print("Protected version: {}".format(version))
			protected.append(version)
			copy_version(version)
	
	if ALWAYS_PATCH_COPY:
		for version in versions:
			if version not in protected:
				print("Copying version: {}".format(version))
				protected.append(version)
				copy_version(version)

	return protected

def patch_profiles(versions):
	profiles_path = os.path.join(MINECRAFT_ROOT, "launcher_profiles.json")
	shutil.copy2(profiles_path, profiles_path + BAK_EXT)

	with open(profiles_path, 'r') as fh:
		launcher_cfg = json.load(fh)

	with open(os.path.join(MINECRAFT_ROOT, "versions/version_manifest_v2.json"), 'r') as fh:
		latests = json.load(fh)['latest']

	add_profiles = {}
	for profile in launcher_cfg['profiles'].values():
		ver = profile['lastVersionId']
		if ver == 'latest-snapshot':
			ver = latests['snapshot']
		elif ver == 'latest-release':
			ver = latests['release']

		if ver in versions:
			print("Updating profile {}({})".format(profile['name'], profile['type']))
			
			if profile['type'] == 'latest-release' or profile['type'] == 'latest-snapshot':
				# Make sure unpatched profile is at the bottom of the list
				old_last_used = profile['lastUsed']
				profile['lastUsed'] = "1970-01-01T00:00:00.000Z"
				
				profile_id = secrets.token_hex(16)
				profile = profile.copy()
				add_profiles[profile_id] = profile
				profile['name'] = "{} (PATCHED)".format(profile['type'])
				profile['type'] = 'custom'
				profile['lastUsed'] = old_last_used
			
			profile['lastVersionId'] = ver + "-patched"
	
	launcher_cfg['profiles'].update(add_profiles)

	with open(profiles_path, 'w') as fh:
		json.dump(launcher_cfg, fh, indent="\t")

def main():
	global MINECRAFT_ROOT

	platform_name = platform.system()
	if platform_name != "Linux":
		print(bold("Warning:"), "Running on untested platform", platform_name)
		print("This script has only been tested on Linux.")
		print("It *should* work on other platforms, but I can't be certain.")
		print("Read all prompts carefully. (Or press ^C to abort.)")
		print()

		if platform_name == "Windows":
			MINECRAFT_ROOT = os.path.expandvars("%APPDATA%/.minecraft")
		elif platform_name == "Darwin":
			MINECRAFT_ROOT = os.path.expanduser("~/Library/Application Support/minecraft")

	print("Press Enter to accept the default value of a prompt.")
	print()

	MINECRAFT_ROOT = prompt_value("Minecraft data directory", MINECRAFT_ROOT)
	print()

	versions = get_mc_versions()
	selected_versions = versions[:]

	while ENABLE_SELECTION:
		print(bold("Minecraft versions in use:"))
		for i, version in enumerate(versions):
			print("({0}) {2} {1}".format(i, version, bold('*') if version in selected_versions else '-'))
		print()
		print("Versions marked with a * will be patched.")
		print("Type the number corresponding to a version to toggle it.")
		print("Or press Enter to patch the selected versions.")

		command = prompt_value("Toggle version", "patch")
		print()
		if command == 'quit':
			print("Abort.")
			assert False
		elif command == 'patch':
			break
		elif command == 'all':
			selected_versions = versions[:]
		else:
			try:
				i = int(command)
				assert i >= 0 and i < len(versions)
				if versions[i] in selected_versions:
					selected_versions.remove(versions[i])
				else:
					selected_versions.append(versions[i])
			except (ValueError, AssertionError):
				pass

	protected_versions = check_manifest(selected_versions)
	selected_versions = [(v+"-patched" if v in protected_versions else v) for v in selected_versions]

	for version in selected_versions:
		try:
			print(bold("Patching version {}...".format(version)))

			deps = get_dependencies(version)
			
			changes = []
			paths = {}
			for dep_name in LOG4J_REPLACE.keys():
				if dep_name in deps:
					lib_name = "{}:{}".format(dep_name, deps[dep_name]['version'])
					print("Uses {}".format(lib_name))
					print(" --> {}".format(deps[dep_name]['path']))
					
					new_path = patch_jar(deps[dep_name]['path'], dep_name)
					
					changes.append(lib_name)
					paths[lib_name] = new_path

			patch_deps(version, changes, paths)

			print()
		except ValueError as ex:
			print(bold("FAILED:"), str(ex))
			print("You may need to restore some .bak'd files manually.")
			print()
			if not prompt_yn("Continue?"):
				break

	patch_profiles(protected_versions)


if __name__ == '__main__':
	main()
